theory Stream_Lib
imports "../Stream/Stream_Behavior_BNF" "../Stream/Stream_More_Corec_Upto5"
begin

(* todo: make them defs eventually: *)
type_synonym stream = J
abbreviation "head xs \<equiv> fst (dtor_J xs)"
abbreviation "tail xs \<equiv> snd (dtor_J xs)"
abbreviation "SCons n xs \<equiv> ctor_J (n, xs)"
code_datatype ctor_J
declare J.dtor_ctor[code]

definition smap where
  "smap f = corec_J (\<lambda>xs. (f (head xs), Inr (tail xs)))"

lemma head_smap[simp]: "head (smap f xs) = f (head xs)"
  unfolding smap_def corec_J_def J.dtor_corec BNF_Comp.id_bnf_comp_def map_pre_J_def by simp

lemma tail_smap[simp]: "tail (smap f xs) = smap f (tail xs)"
  unfolding smap_def corec_J_def J.dtor_corec BNF_Comp.id_bnf_comp_def map_pre_J_def by simp

lemma smap_code[code]: "smap f xs = SCons (f (head xs)) (smap f (tail xs))"
  by (metis J.ctor_dtor prod.collapse head_smap tail_smap)

definition same where
  "same x = corec_J (\<lambda>x. (x, Inr x)) x"

lemma head_same[simp]: "head (same x) = x"
  unfolding same_def corec_J_def J.dtor_corec BNF_Comp.id_bnf_comp_def map_pre_J_def by simp

lemma tail_same[simp]: "tail (same x) = same x"
  unfolding same_def corec_J_def J.dtor_corec BNF_Comp.id_bnf_comp_def map_pre_J_def by simp

definition "sconst x = ctor_J (x, same 0)"

lemma head_sconst[simp]: "head (sconst x) = x"
  unfolding sconst_def by (simp add: J.dtor_ctor)

lemma tail_sconst[simp]: "tail (sconst x) = same 0"
  unfolding sconst_def by (simp add: J.dtor_ctor)

section {* Abbreviations *}
          
abbreviation "SCONS0 \<equiv> gg0  :: 'a \<Sigma>\<Sigma>0 F \<Rightarrow> 'a \<Sigma>\<Sigma>0"
abbreviation "LEAF0 \<equiv> leaf0"
abbreviation GUARD0 :: "'a F \<Rightarrow> 'a F \<Sigma>\<Sigma>0" where "GUARD0 \<equiv> LEAF0"
(* local ("inside") end and continuation: within the upto-corecursive definition of a function f,
END immediately an element of J, while CONT calls f corecursively  *)
abbreviation END0 :: "J \<Rightarrow> (J + 'a) \<Sigma>\<Sigma>0" where "END0 xs \<equiv> LEAF0 (Inl xs)"
abbreviation CONT0 :: "'a \<Rightarrow> (J + 'a) \<Sigma>\<Sigma>0" where "CONT0 a \<equiv> LEAF0 (Inr a)"

abbreviation "SCONS1 \<equiv> gg1  :: 'a \<Sigma>\<Sigma>1 F \<Rightarrow> 'a \<Sigma>\<Sigma>1"
abbreviation "LEAF1 \<equiv> leaf1"
abbreviation GUARD1 :: "'a F \<Rightarrow> 'a F \<Sigma>\<Sigma>1" where "GUARD1 \<equiv> LEAF1"
(* local ("inside") end and continuation: within the upto-corecursive definition of a function f,
END immediately an element of J, while CONT calls f corecursively  *)
abbreviation END1 :: "J \<Rightarrow> (J + 'a) \<Sigma>\<Sigma>1" where "END1 xs \<equiv> LEAF1 (Inl xs)"
abbreviation CONT1 :: "'a \<Rightarrow> (J + 'a) \<Sigma>\<Sigma>1" where "CONT1 a \<equiv> LEAF1 (Inr a)"

abbreviation "SCONS2 \<equiv> gg2  :: 'a \<Sigma>\<Sigma>2 F \<Rightarrow> 'a \<Sigma>\<Sigma>2"
abbreviation "LEAF2 \<equiv> leaf2"
abbreviation GUARD2 :: "'a F \<Rightarrow> 'a F \<Sigma>\<Sigma>2" where "GUARD2 \<equiv> LEAF2"
(* local ("inside") end and continuation: within the upto-corecursive definition of a function f,
END immediately an element of J, while CONT calls f corecursively  *)
abbreviation END2 :: "J \<Rightarrow> (J + 'a) \<Sigma>\<Sigma>2" where "END2 xs \<equiv> LEAF2 (Inl xs)"
abbreviation CONT2 :: "'a \<Rightarrow> (J + 'a) \<Sigma>\<Sigma>2" where "CONT2 a \<equiv> LEAF2 (Inr a)"

abbreviation "SCONS3 \<equiv> gg3  :: 'a \<Sigma>\<Sigma>3 F \<Rightarrow> 'a \<Sigma>\<Sigma>3"
abbreviation "LEAF3 \<equiv> leaf3"
abbreviation GUARD3 :: "'a F \<Rightarrow> 'a F \<Sigma>\<Sigma>3" where "GUARD3 \<equiv> LEAF3"
(* local ("inside") end and continuation: within the upto-corecursive definition of a function f,
END immediately an element of J, while CONT calls f corecursively  *)
abbreviation END3 :: "J \<Rightarrow> (J + 'a) \<Sigma>\<Sigma>3" where "END3 xs \<equiv> LEAF3 (Inl xs)"
abbreviation CONT3 :: "'a \<Rightarrow> (J + 'a) \<Sigma>\<Sigma>3" where "CONT3 a \<equiv> LEAF3 (Inr a)"

abbreviation "SCONS4 \<equiv> gg4  :: 'a \<Sigma>\<Sigma>4 F \<Rightarrow> 'a \<Sigma>\<Sigma>4"
abbreviation "LEAF4 \<equiv> leaf4"
abbreviation GUARD4 :: "'a F \<Rightarrow> 'a F \<Sigma>\<Sigma>4" where "GUARD4 \<equiv> LEAF4"
(* local ("inside") end and continuation: within the upto-corecursive definition of a function f,
END immediately an element of J, while CONT calls f corecursively  *)
abbreviation END4 :: "J \<Rightarrow> (J + 'a) \<Sigma>\<Sigma>4" where "END4 xs \<equiv> LEAF4 (Inl xs)"
abbreviation CONT4 :: "'a \<Rightarrow> (J + 'a) \<Sigma>\<Sigma>4" where "CONT4 a \<equiv> LEAF4 (Inr a)"

abbreviation "SCONS5 \<equiv> gg5  :: 'a \<Sigma>\<Sigma>5 F \<Rightarrow> 'a \<Sigma>\<Sigma>5"
abbreviation "LEAF5 \<equiv> leaf5"
abbreviation GUARD5 :: "'a F \<Rightarrow> 'a F \<Sigma>\<Sigma>5" where "GUARD5 \<equiv> LEAF5"
(* local ("inside") end and continuation: within the upto-corecursive definition of a function f,
END immediately an element of J, while CONT calls f corecursively  *)
abbreviation END5 :: "J \<Rightarrow> (J + 'a) \<Sigma>\<Sigma>5" where "END5 xs \<equiv> LEAF5 (Inl xs)"
abbreviation CONT5 :: "'a \<Rightarrow> (J + 'a) \<Sigma>\<Sigma>5" where "CONT5 a \<equiv> LEAF5 (Inr a)"

end